from enum import IntEnum, StrEnum, unique

__all__ = (
    "ChatAction",
    "ChatMemberStatus",
    "ChatType",
    "ContentType",
    "DiceEmoji",
    "IconColor",
    "MessageEntityType",
    "ParseMode",
    "PollType",
    "RequestMethod",
    "StickerFormat",
    "StickerType",
    "UpdateType",
)


@unique
class RequestMethod(StrEnum):
    GET = "GET"
    POST = "POST"


@unique
class ChatType(StrEnum):
    PRIVATE = "private"
    GROUP = "group"
    SUPERGROUP = "supergroup"
    CHANNEL = "channel"


@unique
class ChatAction(StrEnum):
    TYPING = "typing"
    UPLOAD_PHOTO = "upload_photo"
    RECORD_VIDEO = "record_video"
    UPLOAD_VIDEO = "upload_video"
    RECORD_VOICE = "record_voice"
    UPLOAD_VOICE = "upload_voice"
    UPLOAD_DOCUMENT = "upload_document"
    CHOOSE_STICKER = "choose_sticker"
    FIND_LOCATION = "find_location"
    RECORD_VIDEO_NOTE = "record_video_note"
    UPLOAD_VIDEO_NOTE = "upload_video_note"


@unique
class ChatMemberStatus(StrEnum):
    CREATOR = "creator"
    ADMINISTRATOR = "administrator"
    MEMBER = "member"
    RESTRICTED = "restricted"
    LEFT = "left"
    KICKED = "kicked"


@unique
class MessageEntityType(StrEnum):
    MENTION = "mention"
    HASHTAG = "hashtag"
    CASHTAG = "cashtag"
    BOT_COMMAND = "bot_command"
    URL = "url"
    EMAIL = "email"
    PHONE_NUMBER = "phone_number"
    BOLD = "bold"
    ITALIC = "italic"
    CODE = "code"
    PRE = "pre"
    TEXT_LINK = "text_link"
    TEXT_MENTION = "text_mention"
    UNDERLINE = "underline"
    STRIKETHROUGH = "strikethrough"
    SPOILER = "spoiler"
    CUSTOM_EMOJI = "custom_emoji"


@unique
class UpdateType(StrEnum):
    MESSAGE = "message"
    EDITED_MESSAGE = "edited_message"
    CHANNEL_POST = "channel_post"
    EDITED_CHANNEL_POST = "edited_channel_post"
    INLINE_QUERY = "inline_query"
    CHOSEN_INLINE_RESULT = "chosen_inline_result"
    CALLBACK_QUERY = "callback_query"
    SHIPPING_QUERY = "shipping_query"
    PRE_CHECKOUT_QUERY = "pre_checkout_query"
    POLL = "poll"
    POLL_ANSWER = "poll_answer"
    MY_CHAT_MEMBER = "my_chat_member"
    CHAT_MEMBER = "chat_member"


@unique
class ContentType(StrEnum):
    TEXT = "text"
    AUDIO = "audio"
    DOCUMENT = "document"
    ANIMATION = "animation"
    GAME = "game"
    PHOTO = "photo"
    STICKER = "sticker"
    VIDEO = "video"
    VIDEO_NOTE = "video_note"
    VOICE = "voice"
    CONTACT = "contact"
    LOCATION = "location"
    VENUE = "venue"
    NEW_CHAT_MEMBERS = "new_chat_members"
    LEFT_CHAT_MEMBER = "left_chat_member"
    INVOICE = "invoice"
    SUCCESSFUL_PAYMENT = "successful_payment"
    CONNECTED_WEBSITE = "connected_website"
    MIGRATE_TO_CHAT_ID = "migrate_to_chat_id"
    MIGRATE_FROM_CHAT_ID = "migrate_from_chat_id"
    PINNED_MESSAGE = "pinned_message"
    NEW_CHAT_TITLE = "new_chat_title"
    NEW_CHAT_PHOTO = "new_chat_photo"
    DELETE_CHAT_PHOTO = "delete_chat_photo"
    GROUP_CHAT_CREATED = "group_chat_created"
    PASSPORT_DATA = "passport_data"


@unique
class ParseMode(StrEnum):
    MARKDOWN = "Markdown"
    HTML = "HTML"
    MARKDOWNV2 = "MarkdownV2"


@unique
class PollType(StrEnum):
    QUIZ = "quiz"
    REGULAR = "regular"


@unique
class DiceEmoji(StrEnum):
    DICE = "🎲"
    DARTS = "🎯"
    BASKETBALL = "🏀"
    FOOTBALL = "⚽"
    SLOT_MACHINE = "🎰"
    BOWLING = "🎳"


@unique
class StickerFormat(StrEnum):
    STATIC = "static"
    ANIMATED = "animated"
    VIDEO = "video"


@unique
class StickerType(StrEnum):
    REGULAR = "regular"
    MASK = "mask"
    CUSTOM_EMOJI = "custom_emoji"


class MaskPositionPoint(StrEnum):
    FOREHEAD = "forehead"
    EYES = "eyes"
    MOUTH = "mouth"
    CHIN = "chin"


class IconColor(IntEnum):
    IC_0x6FB9F0 = 0x6FB9F0
    IC_0xFFD67E = 0xFFD67E
    IC_0xCB86DB = 0xCB86DB
    IC_0x8EEE98 = 0x8EEE98
    IC_0xFF93B2 = 0xFF93B2
    IC_0xFB6F5F = 0xFB6F5F



from typing import Any, AsyncIterator, Final

from .helpers import Json
from .storage import StorageProtocol

__all__ = ("MemoryStorage",)


class MemoryStorage(StorageProtocol):
    def __init__(self) -> None:
        self._data: Final[dict[str, Any]] = {}

    async def connect(self) -> None: ...

    async def close(self) -> None: ...

    async def set(self, key: str, value: Json = None) -> None:
        self._data[key] = value

    async def get(self, key: str) -> Json:
        return self._data.get(key)

    async def delete(self, key: str) -> None:
        self._data.pop(key)

    async def iterate(
        self, prefix: str = ""
    ) -> AsyncIterator[tuple[str, Json]]:
        for key, value in self._data.items():
            if key.startswith(prefix):
                yield key, value

    async def clear(self) -> None:
        self._data.clear()

    def raw_connection(self) -> Any:
        return None



__version__ = "0.17.0"

from .api_types import (
    API,
    CallbackQuery,
    Chat,
    ChosenInlineResult,
    Contact,
    File,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    InlineQuery,
    KeyboardButton,
    LocalFile,
    Message,
    PreCheckoutQuery,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    ShippingQuery,
    StreamFile,
    User,
)
from .bot import Bot, FilterProtocol, PollBot
from .bot_update import BotUpdate
from .constants import (
    ChatAction,
    ChatMemberStatus,
    ChatType,
    ContentType,
    ParseMode,
    PollType,
    UpdateType,
)
from .exceptions import (
    BadGateway,
    BotBlocked,
    BotKicked,
    MigrateToChat,
    RestartingTelegram,
    RetryAfter,
    TelegramError,
)
from .filters import (
    ANDFilter,
    CallbackQueryDataFilter,
    CommandsFilter,
    ContentTypeFilter,
    GroupChatFilter,
    MessageTextFilter,
    ORFilter,
    PrivateChatFilter,
    StateFilter,
    UpdateTypeFilter,
)
from .handler_table import HandlerTable
from .storage import StorageProtocol

__all__ = (
    "ANDFilter",
    "API",
    "BadGateway",
    "Bot",
    "BotBlocked",
    "BotKicked",
    "BotUpdate",
    "CallbackQuery",
    "CallbackQueryDataFilter",
    "Chat",
    "ChatAction",
    "ChatMemberStatus",
    "ChatType",
    "ChosenInlineResult",
    "CommandsFilter",
    "Contact",
    "ContentType",
    "ContentTypeFilter",
    "File",
    "FilterProtocol",
    "GroupChatFilter",
    "HandlerTable",
    "InlineKeyboardButton",
    "InlineKeyboardMarkup",
    "InlineQuery",
    "KeyboardButton",
    "LocalFile",
    "Message",
    "MessageTextFilter",
    "MigrateToChat",
    "ORFilter",
    "ParseMode",
    "PollBot",
    "PollType",
    "PreCheckoutQuery",
    "PrivateChatFilter",
    "ReplyKeyboardMarkup",
    "ReplyKeyboardRemove",
    "RestartingTelegram",
    "RetryAfter",
    "ShippingQuery",
    "StateFilter",
    "StorageProtocol",
    "StreamFile",
    "TelegramError",
    "UpdateType",
    "UpdateTypeFilter",
    "User",
    "__version__",
)



from base64 import b64decode
from typing import AsyncIterator, Final, Self, Sequence, Union, cast

import msgspec.json
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.padding import MGF1, OAEP
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.ciphers import Cipher, CipherContext
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.constant_time import bytes_eq
from cryptography.hazmat.primitives.hashes import SHA1, SHA256, SHA512, Hash
from yarl import URL

from .api_types import API, EncryptedCredentials, PassportElementType

__all__ = (
    "Credentials",
    "DataCredentials",
    "FileCredentials",
    "IdDocumentData",
    "PassportCipher",
    "PassportKey",
    "PassportScope",
    "PassportScopeElement",
    "PassportScopeElementOne",
    "PassportScopeElementOneOfSeveral",
    "PersonalDetails",
    "ResidentialAddress",
    "SecureData",
    "SecureValue",
    "passport_request",
)


def passport_request(
    bot_id: int,
    scope: "PassportScope",
    public_key: str,
    nonce: str,
) -> str:
    url = URL("tg://resolve").with_query(
        domain="telegrampassport",
        bot_id=bot_id,
        scope=msgspec.json.encode(scope).decode(),
        public_key=public_key,
        nonce=nonce,
    )
    return str(url)


class PassportKey:
    _padding: Final[OAEP] = OAEP(
        mgf=MGF1(algorithm=SHA1()),
        algorithm=SHA1(),
        label=None,
    )

    def __init__(
        self,
        private_key: RSAPrivateKey,
    ) -> None:
        if not isinstance(private_key, RSAPrivateKey):
            raise RuntimeError("Key is not RSA private key")
        self._private_key: Final[RSAPrivateKey] = private_key
        public_key = self._private_key.public_key()
        public_bytes = public_key.public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        self._public_key_pem: Final[str] = public_bytes.decode()

    @classmethod
    def load_der(
        cls,
        private_bytes: bytes,
    ) -> Self:
        private_key = serialization.load_der_private_key(
            private_bytes,
            password=None,
        )
        return cls(cast(RSAPrivateKey, private_key))

    @classmethod
    def load_pem(
        cls,
        private_text: str,
    ) -> Self:
        private_key = serialization.load_pem_private_key(
            private_text.encode(),
            password=None,
        )
        return cls(cast(RSAPrivateKey, private_key))

    def decrypt(
        self,
        ciphertext: bytes,
    ) -> bytes:
        return self._private_key.decrypt(ciphertext, self._padding)

    @property
    def public_key_pem(
        self,
    ) -> str:
        return self._public_key_pem


class PassportCipher:
    _key_size: Final[int] = 32
    _iv_size: Final[int] = 16

    def __init__(
        self,
        data_secret: bytes,
        data_hash: bytes,
    ) -> None:
        digest = Hash(SHA512())
        digest.update(data_secret)
        digest.update(data_hash)
        secret_hash = digest.finalize()
        key = secret_hash[: self._key_size]
        iv = secret_hash[
            self._key_size : self._key_size + self._iv_size  # noqa: E203
        ]
        self._data_hash: Final[bytes] = data_hash
        self._cipher: Final[Cipher[CBC]] = Cipher(AES(key), CBC(iv))

    def decrypt(
        self,
        ciphertext: bytes,
    ) -> bytes:
        decryptor = self._cipher.decryptor()
        assert isinstance(decryptor, CipherContext)
        plaintext = decryptor.update(ciphertext) + decryptor.finalize()
        digest = Hash(SHA256())
        digest.update(plaintext)
        computed_hash = digest.finalize()
        if not bytes_eq(computed_hash, self._data_hash):
            raise RuntimeError("Decryption error")
        return plaintext[plaintext[0] :]  # noqa: E203

    async def decrypt_stream(
        self,
        stream: AsyncIterator[bytes],
    ) -> AsyncIterator[bytes]:
        decryptor = self._cipher.decryptor()
        assert isinstance(decryptor, CipherContext)
        digest = Hash(SHA256())
        skip = None
        async for chunk in stream:
            decrypted = decryptor.update(chunk)
            digest.update(decrypted)
            if skip is None:
                skip = decrypted[0]
            if skip >= len(decrypted):
                skip = skip - len(decrypted)
            else:
                yield decrypted[skip:]
                skip = 0
        decrypted = decryptor.finalize()
        digest.update(decrypted)
        computed_hash = digest.finalize()
        if not bytes_eq(computed_hash, self._data_hash):
            raise RuntimeError("Decryption error")
        yield decrypted[skip:]


class PassportScopeElementOne(API, frozen=True, kw_only=True):
    type: PassportElementType
    selfie: bool | None = None
    translation: bool | None = None
    native_names: bool | None = None


class PassportScopeElementOneOfSeveral(API, frozen=True, kw_only=True):
    one_of: Sequence[PassportScopeElementOne]
    selfie: bool | None = None
    translation: bool | None = None


PassportScopeElement = Union[
    PassportScopeElementOne,
    PassportScopeElementOneOfSeveral,
]


class PassportScope(
    API,
    frozen=True,
    tag_field="v",
    tag=1,
    kw_only=True,
):
    data: Sequence[PassportScopeElement]


class FileCredentials(API, frozen=True, kw_only=True):
    file_hash: str
    secret: str


class DataCredentials(API, frozen=True, kw_only=True):
    data_hash: str
    secret: str

    def decrypt(
        self,
        ciphertext: str,
    ) -> bytes:
        cipher = PassportCipher(
            b64decode(self.secret),
            b64decode(self.data_hash),
        )
        return cipher.decrypt(b64decode(ciphertext))


class SecureValue(API, frozen=True, kw_only=True):
    data: DataCredentials | None = None
    front_side: FileCredentials | None = None
    reverse_side: FileCredentials | None = None
    selfie: FileCredentials | None = None
    translation: Sequence[FileCredentials] | None = None
    files: Sequence[FileCredentials] | None = None


class SecureData(API, frozen=True, kw_only=True):
    personal_details: SecureValue | None = None
    passport: SecureValue | None = None
    internal_passport: SecureValue | None = None
    driver_license: SecureValue | None = None
    identity_card: SecureValue | None = None
    address: SecureValue | None = None
    utility_bill: SecureValue | None = None
    bank_statement: SecureValue | None = None
    rental_agreement: SecureValue | None = None
    passport_registration: SecureValue | None = None
    temporary_registration: SecureValue | None = None


class Credentials(API, frozen=True, kw_only=True):
    secure_data: SecureData
    nonce: str

    @classmethod
    def from_encrypted(
        cls,
        encrypted: EncryptedCredentials,
        passport_key: PassportKey,
    ) -> Self:
        data_secret = passport_key.decrypt(b64decode(encrypted.secret))
        data_hash = b64decode(encrypted.hash)
        ciphertext = b64decode(encrypted.data)
        cipher = PassportCipher(data_secret, data_hash)
        plaintext = cipher.decrypt(ciphertext)
        return msgspec.json.decode(plaintext, type=cls)


class PersonalDetails(API, frozen=True, kw_only=True):
    first_name: str
    last_name: str
    birth_date: str
    gender: str
    country_code: str
    residence_country_code: str
    middle_name: str | None = None
    first_name_native: str | None = None
    last_name_native: str | None = None
    middle_name_native: str | None = None


class ResidentialAddress(API, frozen=True, kw_only=True):
    street_line1: str
    city: str
    country_code: str
    post_code: str
    street_line2: str | None = None
    state: str | None = None


class IdDocumentData(API, frozen=True, kw_only=True):
    document_no: str
    expiry_date: str | None = None



import asyncio
import logging
from ipaddress import IPv4Address, IPv4Network
from secrets import compare_digest, token_urlsafe
from typing import Any, Final

import msgspec.json
from aiohttp import ClientSession
from aiohttp.web import (
    Application,
    HTTPInternalServerError,
    HTTPNotFound,
    Request,
    Response,
    StreamResponse,
)
from yarl import URL

from .api_types import InputFile, Update
from .bot import Bot, HandlerTableProtocol
from .storage import StorageProtocol

NETWORKS: Final[tuple[IPv4Network, ...]] = (
    IPv4Network("149.154.160.0/20"),
    IPv4Network("91.108.4.0/22"),
)

bot_logger: Final[logging.Logger] = logging.getLogger("aiotgbot.bot")


class ListenBot(Bot):
    def __init__(
        self,
        url: str | URL,
        token: str,
        handler_table: "HandlerTableProtocol",
        storage: StorageProtocol,
        certificate: InputFile | None = None,
        ip_address: str | None = None,
        check_address: bool = False,
        address_header: str | None = None,
        client_session: ClientSession | None = None,
        **application_args: Any
    ) -> None:
        super().__init__(
            token,
            handler_table,
            storage,
            client_session,
        )
        self._url: URL = URL(url) if isinstance(url, str) else url
        self._certificate = certificate
        self._ip_address = ip_address
        self._webhook_token: str | None = None
        self._check_address: Final[bool] = check_address
        self._address_header: Final[str | None] = address_header
        self._application = Application(**application_args)
        self._application.router.add_post("/{token}", self._handler)

    @property
    def application(self) -> Application:
        return self._application

    def _address_is_allowed(self, request: Request) -> bool:
        if self._address_header is not None:
            address = IPv4Address(request.headers[self._address_header])
        else:
            address = IPv4Address(request.remote)
        return any(address in network for network in NETWORKS)

    async def _handler(self, request: Request) -> StreamResponse:
        if not self._started:
            raise HTTPInternalServerError()
        assert self._scheduler is not None
        assert self._webhook_token is not None
        if self._check_address and not self._address_is_allowed(request):
            raise HTTPNotFound()
        if not compare_digest(
            self._webhook_token, request.match_info["token"]
        ):
            raise HTTPNotFound()
        update_data = await request.read()
        update = msgspec.json.decode(update_data, type=Update)
        await self._scheduler.spawn(self._handle_update(update))
        return Response()

    async def start(self) -> None:
        if self._started:
            raise RuntimeError("Polling already started")
        await self._start()
        assert self._me is not None
        loop = asyncio.get_running_loop()
        self._webhook_token = await loop.run_in_executor(None, token_urlsafe)
        assert isinstance(self._webhook_token, str)
        url = str(self._url / self._webhook_token)
        await self.set_webhook(url, self._certificate, self._ip_address)
        bot_logger.info(
            "Bot %s (%s) start listen",
            self._me.first_name,
            self._me.username,
        )

    async def stop(self) -> None:
        if not self._started:
            raise RuntimeError("Polling not started")
        if self._stopped:
            raise RuntimeError("Polling already stopped")
        assert self._me is not None
        self._stopped = True
        await self.delete_webhook()
        await self._cleanup()
        bot_logger.info(
            "Bot %s (%s) stop listen",
            self._me.first_name,
            self._me.username,
        )



from typing import Any, Final, Iterator, MutableMapping

from .api_types import (
    CallbackQuery,
    ChatMemberUpdated,
    ChosenInlineResult,
    InlineQuery,
    Message,
    Poll,
    PollAnswer,
    PreCheckoutQuery,
    ShippingQuery,
    Update,
)

__all__ = (
    "BotUpdate",
    "Context",
)


class Context(MutableMapping[str, Any]):
    def __init__(self, data: dict[str, Any]) -> None:
        self._data: Final[dict[str, Any]] = data

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def clear(self) -> None:
        self._data.clear()

    def to_dict(self) -> dict[str, Any]:
        return self._data


class BotUpdate(MutableMapping[str, Any]):
    def __init__(
        self, state: str | None, context: Context, update: Update
    ) -> None:
        self._state: str | None = state
        self._context: Final[Context] = context
        self._update: Final[Update] = update
        self._data: Final[dict[str, Any]] = {}

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    @property
    def state(self) -> str | None:
        return self._state

    @state.setter
    def state(self, value: str) -> None:
        self._state = value

    @property
    def context(self) -> Context:
        return self._context

    @property
    def update_id(self) -> int:
        return self._update.update_id

    @property
    def message(self) -> Message | None:
        return self._update.message

    @property
    def edited_message(self) -> Message | None:
        return self._update.edited_message

    @property
    def channel_post(self) -> Message | None:
        return self._update.channel_post

    @property
    def edited_channel_post(self) -> Message | None:
        return self._update.edited_channel_post

    @property
    def inline_query(self) -> InlineQuery | None:
        return self._update.inline_query

    @property
    def chosen_inline_result(self) -> ChosenInlineResult | None:
        return self._update.chosen_inline_result

    @property
    def callback_query(self) -> CallbackQuery | None:
        return self._update.callback_query

    @property
    def shipping_query(self) -> ShippingQuery | None:
        return self._update.shipping_query

    @property
    def pre_checkout_query(self) -> PreCheckoutQuery | None:
        return self._update.pre_checkout_query

    @property
    def poll(self) -> Poll | None:
        return self._update.poll

    @property
    def poll_answer(self) -> PollAnswer | None:
        return self._update.poll_answer

    @property
    def my_chat_member(self) -> ChatMemberUpdated | None:
        return self._update.my_chat_member

    @property
    def chat_member(self) -> ChatMemberUpdated | None:
        return self._update.chat_member



import asyncio
import logging
from abc import abstractmethod
from dataclasses import dataclass
from functools import partial
from http import HTTPStatus
from typing import (
    Any,
    Awaitable,
    Callable,
    Final,
    Iterator,
    MutableMapping,
    Protocol,
    Type,
    TypeVar,
    overload,
    runtime_checkable,
)

import aiojobs
import backoff
import msgspec
from aiofreqlimit import FreqLimit
from aiohttp import ClientError, ClientSession, FormData, TCPConnector

from .api_methods import ApiMethods, ParamType
from .api_types import APIResponse, InputFile, Update, User
from .bot_update import BotUpdate, Context
from .constants import ChatType, RequestMethod
from .exceptions import (
    BadGateway,
    BotBlocked,
    BotKicked,
    MigrateToChat,
    RestartingTelegram,
    RetryAfter,
    TelegramError,
)
from .helpers import BotKey, KeyLock, get_software
from .storage import StorageProtocol

__all__ = (
    "Bot",
    "FilterProtocol",
    "Handler",
    "HandlerCallable",
    "HandlerTableProtocol",
    "PollBot",
)

SOFTWARE: Final[str] = get_software()
TG_API_URL: Final[str] = "https://api.telegram.org/bot{token}/{method}"
TG_FILE_URL: Final[str] = "https://api.telegram.org/file/bot{token}/{path}"
TG_GET_UPDATES_TIMEOUT: Final[int] = 60
STATE_PREFIX: Final[str] = "state"
CONTEXT_PREFIX: Final[str] = "context"
MESSAGE_INTERVAL: Final[float] = 1 / 30
CHAT_INTERVAL: Final[float] = 1
GROUP_INTERVAL: Final[float] = 3

bot_logger: Final[logging.Logger] = logging.getLogger("aiotgbot.bot")
response_logger: Final[logging.Logger] = logging.getLogger("aiotgbot.response")

EventHandler = Callable[["Bot"], Awaitable[None]]

T = TypeVar("T")
V = TypeVar("V")


class Bot(MutableMapping[str | BotKey[Any], Any], ApiMethods):
    def __init__(
        self,
        token: str,
        handler_table: "HandlerTableProtocol",
        storage: StorageProtocol,
        client_session: ClientSession | None = None,
    ) -> None:
        if not handler_table.frozen:
            raise RuntimeError("Can't use unfrozen handler table")
        if not isinstance(handler_table, HandlerTableProtocol):
            raise RuntimeError(
                "Handler table is not HandlerTableProtocol instance"
            )
        if not isinstance(storage, StorageProtocol):
            raise RuntimeError(
                "Storage parameter is not StorageProtocol instance"
            )
        self._token: Final[str] = token
        self._handler_table: Final["HandlerTableProtocol"] = handler_table
        self._storage: Final[StorageProtocol] = storage
        if client_session is not None:
            client_session.headers.setdefault("User-Agent", SOFTWARE)
        else:
            client_session = ClientSession(
                connector=TCPConnector(keepalive_timeout=60),
                headers={"User-Agent": SOFTWARE},
            )
        self._client_session: Final[ClientSession] = client_session
        self._context_lock: Final[KeyLock] = KeyLock()
        self._message_limit: Final[FreqLimit] = FreqLimit(MESSAGE_INTERVAL)
        self._chat_limit: Final[FreqLimit] = FreqLimit(CHAT_INTERVAL)
        self._group_limit: Final[FreqLimit] = FreqLimit(GROUP_INTERVAL)
        self._scheduler: aiojobs.Scheduler | None = None
        self._started: bool = False
        self._stopped: bool = False
        self._updates_offset: int = 0
        self._me: User | None = None
        self._data: Final[dict[BotKey[Any] | str, object]] = {}

    @overload  # type: ignore[override]
    def __getitem__(self, key: BotKey[T]) -> T: ...

    @overload
    def __getitem__(self, key: str) -> Any: ...

    def __getitem__(self, key: str | BotKey[T]) -> Any:
        return self._data[key]

    @overload  # type: ignore[override]
    def __setitem__(self, key: BotKey[T], value: T) -> None: ...

    @overload
    def __setitem__(self, key: str, value: Any) -> None: ...

    def __setitem__(self, key: str | BotKey[T], value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str | BotKey[T]) -> None:
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str | BotKey[Any]]:
        return iter(self._data)

    @property
    def id(self) -> int:
        return int(self._token.split(":")[0])

    @property
    def storage(self) -> StorageProtocol:
        return self._storage

    @property
    def client(self) -> ClientSession:
        if self._client_session is None:
            raise RuntimeError("Access to client during bot is not running.")
        else:
            return self._client_session

    def file_url(self, path: str) -> str:
        return TG_FILE_URL.format(token=self._token, path=path)

    @staticmethod
    def _scheduler_exception_handler(
        _: aiojobs.Scheduler, context: dict[str, Any]
    ) -> None:
        bot_logger.exception(
            "Update handle error", exc_info=context["exception"]
        )

    @staticmethod
    def _telegram_exception(api_response: APIResponse) -> TelegramError:
        assert api_response.error_code is not None
        assert api_response.description is not None
        error_code = api_response.error_code
        description = api_response.description
        if (
            api_response.parameters is not None
            and api_response.parameters.retry_after is not None
        ):
            retry_after = api_response.parameters.retry_after
            assert retry_after is not None
            return RetryAfter(error_code, description, retry_after)
        elif (
            api_response.parameters is not None
            and api_response.parameters.migrate_to_chat_id is not None
        ):
            return MigrateToChat(
                error_code,
                description,
                api_response.parameters.migrate_to_chat_id,
            )
        elif (
            error_code >= HTTPStatus.INTERNAL_SERVER_ERROR
            and RestartingTelegram.match(description)
        ):
            return RestartingTelegram(error_code, description)
        elif BadGateway.match(description):
            return BadGateway(error_code, description)
        elif BotBlocked.match(description):
            return BotBlocked(error_code, description)
        elif BotKicked.match(description):
            return BotKicked(error_code, description)
        else:
            return TelegramError(error_code, description)

    @backoff.on_exception(backoff.expo, ClientError)
    async def _request(
        self,
        http_method: RequestMethod,
        api_method: str,
        type_: Type[V],
        **params: ParamType,
    ) -> V:
        data = {
            name: str(value) if isinstance(value, (int, float)) else value
            for name, value in params.items()
            if value is not None
        }
        bot_logger.debug(
            "Request %s %s %r",
            http_method,
            api_method,
            data,
        )
        if http_method == RequestMethod.GET:
            if len(data) > 0:
                assert all(isinstance(value, str) for value in data.values())
                request = partial(self.client.get, params=data)
            else:
                request = partial(self.client.get)
        else:
            form_data = FormData()
            for name, value in data.items():
                if isinstance(value, InputFile):
                    form_data.add_field(
                        name,
                        value.content,
                        content_type=value.content_type,
                        filename=value.name,
                    )
                else:
                    form_data.add_field(name, value)
            request = partial(self.client.post, data=form_data)

        url = TG_API_URL.format(token=self._token, method=api_method)
        async with request(url) as response:
            response_data = await response.read()
        api_response = msgspec.json.decode(response_data, type=APIResponse)
        if api_response.ok:
            assert isinstance(api_response.result, msgspec.Raw)
            return msgspec.json.decode(api_response.result, type=type_)
        else:
            assert api_response.result is msgspec.UNSET
            raise Bot._telegram_exception(api_response)

    async def _safe_request(
        self,
        http_method: RequestMethod,
        api_method: str,
        chat_id: int | str,
        type_: Type[V],
        **params: ParamType,
    ) -> V:
        retry_allowed = all(
            not isinstance(param, InputFile) for param in params.values()
        )
        request: Callable[[], Awaitable[V]] = partial(
            self._request,
            http_method,
            api_method,
            type_,
            chat_id=chat_id,
            **params,
        )

        chat = await self.get_chat(chat_id)
        while True:
            try:
                message_limit = self._message_limit.resource()
                if chat.type in (ChatType.GROUP, ChatType.SUPERGROUP):
                    group_limit = self._group_limit.resource(chat.id)
                    async with message_limit, group_limit:
                        return await request()
                else:
                    chat_limit = self._chat_limit.resource(chat.id)
                    async with message_limit, chat_limit:
                        return await request()
            except RetryAfter as retry_after:
                if retry_allowed:
                    await asyncio.sleep(retry_after.retry_after)
                else:
                    bot_logger.error(
                        "RetryAfter error during retry not allowed",
                    )
                    raise

    @staticmethod
    def _update_state(update: Update) -> str:
        user_id: int | None = None
        chat_id: int | None = None

        if update.message is not None:
            assert update.message.from_ is not None
            user_id = update.message.from_.id
            assert update.message.chat is not None
            chat_id = update.message.chat.id
        elif update.edited_message is not None:
            assert update.edited_message.from_ is not None
            user_id = update.edited_message.from_.id
            assert update.edited_message.chat is not None
            chat_id = update.edited_message.chat.id
        elif update.channel_post is not None:
            assert update.channel_post.chat is not None
            chat_id = update.channel_post.chat.id
        elif update.edited_channel_post is not None:
            assert update.edited_channel_post.chat is not None
            chat_id = update.edited_channel_post.chat.id
        elif update.inline_query is not None:
            user_id = update.inline_query.from_.id
        elif update.chosen_inline_result is not None:
            user_id = update.chosen_inline_result.from_.id
        elif (
            update.callback_query is not None
            and update.callback_query.message is not None
        ):
            user_id = update.callback_query.from_.id
            assert update.callback_query.message.chat is not None
            chat_id = update.callback_query.message.chat.id
        elif update.callback_query is not None:
            user_id = update.callback_query.from_.id
        elif update.shipping_query is not None:
            user_id = update.shipping_query.from_.id
        elif update.pre_checkout_query is not None:
            user_id = update.pre_checkout_query.from_.id
        elif update.poll is not None:
            pass
        elif (
            update.poll_answer is not None
            and update.poll_answer.user is not None
        ):
            user_id = update.poll_answer.user.id
        elif update.my_chat_member is not None:
            assert update.my_chat_member.from_ is not None
            user_id = update.my_chat_member.from_.id
            assert update.my_chat_member.chat is not None
            chat_id = update.my_chat_member.chat.id
        elif update.chat_member is not None:
            assert update.chat_member.from_ is not None
            user_id = update.chat_member.from_.id
            assert update.chat_member.chat is not None
            chat_id = update.chat_member.chat.id

        return "{}|{}".format(
            str(user_id) if user_id is not None else "",
            str(chat_id) if chat_id is not None else "",
        )

    async def _handle_update(self, update: Update) -> None:
        assert self._handler_table.frozen
        assert self._context_lock is not None, "Context lock not initialized"
        bot_logger.debug(
            'Dispatch update "%s"',
            update.update_id,
        )
        update_state = self._update_state(update)
        state_key = f"{STATE_PREFIX}|{update_state}"
        context_key = f"{CONTEXT_PREFIX}|{update_state}"
        async with self._context_lock.resource(state_key):
            state = await self._storage.get(state_key)
            assert isinstance(state, str) or state is None
            context_dict = await self._storage.get(context_key)
            assert isinstance(context_dict, dict) or context_dict is None
            context = Context(context_dict if context_dict is not None else {})
            bot_update = BotUpdate(state, context, update)
            handler = await self._handler_table.get_handler(self, bot_update)
            if handler is not None:
                bot_logger.debug(
                    'Dispatched update "%s" to "%s"',
                    update.update_id,
                    handler.__name__,
                )
                await handler(self, bot_update)
                await self._storage.set(state_key, bot_update.state)
                await self._storage.set(
                    context_key, bot_update.context.to_dict()
                )
                bot_logger.debug(
                    'Set context for update "%s"',
                    update.update_id,
                )
            else:
                bot_logger.debug(
                    'Not found handler for update "%s". Skip.',
                    update.update_id,
                )

    async def _start(self) -> None:
        self._started = True

        self._me = await self.get_me()
        self._scheduler = aiojobs.Scheduler(
            exception_handler=self._scheduler_exception_handler
        )

    async def _cleanup(self) -> None:
        assert self._client_session is not None
        assert self._scheduler is not None
        await self._scheduler.close()
        await self._client_session.close()
        await self._message_limit.clear()
        await self._chat_limit.clear()
        await self._group_limit.clear()

    @abstractmethod
    async def start(self) -> None: ...

    @abstractmethod
    async def stop(self) -> None: ...


class PollBot(Bot):
    def __init__(
        self,
        token: str,
        handler_table: "HandlerTableProtocol",
        storage: StorageProtocol,
        client_session: ClientSession | None = None,
    ) -> None:
        super().__init__(
            token,
            handler_table,
            storage,
            client_session,
        )
        self._poll_task: asyncio.Task[None] | None = None

    async def start(self) -> None:
        if self._started:
            raise RuntimeError("Polling already started")
        await self._start()
        assert self._me is not None
        self._poll_task = asyncio.create_task(self._poll_wrapper())
        bot_logger.info(
            "Bot %s (%s) start polling",
            self._me.first_name,
            self._me.username,
        )

    async def stop(self) -> None:
        if not self._started:
            raise RuntimeError("Polling not started")
        if self._stopped:
            raise RuntimeError("Polling already stopped")
        assert self._poll_task is not None
        bot_logger.debug("Stop polling")
        self._stopped = True
        if not self._poll_task.done():
            self._poll_task.cancel()

    async def _poll_wrapper(self) -> None:
        assert self._me is not None
        try:
            await self._poll()
        except asyncio.CancelledError:
            pass
        except Exception as exception:
            bot_logger.exception(
                "Error while polling updates",
                exc_info=exception,
            )
        await self._cleanup()
        bot_logger.info(
            "Bot %s (%s) stop polling",
            self._me.first_name,
            self._me.username,
        )

    @backoff.on_exception(backoff.expo, TelegramError)
    async def _poll(self) -> None:
        assert self._scheduler is not None, "Scheduler not initialized"
        bot_logger.debug("Get updates from: %s", self._updates_offset)
        while not self._stopped:
            updates = await self.get_updates(
                offset=self._updates_offset,
                timeout=TG_GET_UPDATES_TIMEOUT,
            )
            for update in updates:
                await self._scheduler.spawn(
                    self._handle_update(update),
                    f"handle_update_{update.update_id}",
                )
            if len(updates) > 0:
                self._updates_offset = updates[-1].update_id + 1


HandlerCallable = Callable[[Bot, BotUpdate], Awaitable[None]]
FiltersType = tuple["FilterProtocol", ...]


@dataclass(frozen=True)
class Handler:
    callable: HandlerCallable
    filters: FiltersType

    async def check(self, bot: Bot, update: BotUpdate) -> bool:
        return all(
            [await _filter.check(bot, update) for _filter in self.filters]
        )


@runtime_checkable
class HandlerTableProtocol(Protocol):
    def freeze(self) -> None: ...

    @property
    @abstractmethod
    def frozen(self) -> bool: ...

    @abstractmethod
    async def get_handler(
        self, bot: Bot, update: BotUpdate
    ) -> HandlerCallable | None: ...


@runtime_checkable
class FilterProtocol(Protocol):
    @abstractmethod
    async def check(self, bot: Bot, update: BotUpdate) -> bool: ...



import logging
from abc import ABC, abstractmethod
from itertools import count
from typing import Any, Final, Iterable, Sequence, Type, TypeVar

import msgspec

from .api_types import (
    BotCommand,
    BotCommandScope,
    BotDescription,
    BotName,
    BotShortDescription,
    Chat,
    ChatAdministratorRights,
    ChatInviteLink,
    ChatMember,
    ChatPermissions,
    File,
    ForumTopic,
    GameHighScore,
    InlineKeyboardMarkup,
    InlineQueryResult,
    InlineQueryResultsButton,
    InputFile,
    InputMedia,
    InputMediaWithThumbnail,
    InputSticker,
    LabeledPrice,
    LinkPreviewOptions,
    MaskPosition,
    MenuButton,
    Message,
    MessageEntity,
    MessageId,
    PassportElementError,
    Poll,
    ReactionType,
    ReplyMarkup,
    ReplyParameters,
    SentWebAppMessage,
    ShippingOption,
    Sticker,
    StickerSet,
    Update,
    User,
    UserChatBoosts,
    UserProfilePhotos,
    WebhookInfo,
)
from .constants import (
    ChatAction,
    DiceEmoji,
    IconColor,
    ParseMode,
    PollType,
    RequestMethod,
    StickerFormat,
    StickerType,
    UpdateType,
)

__all__ = (
    "ApiMethods",
    "ParamType",
)

api_logger: Final[logging.Logger] = logging.getLogger("aiotgbot.api")


ParamType = int | float | str | InputFile | None


def _encode_json(obj: Any) -> str | None:
    if obj is not None:
        return msgspec.json.encode(obj).decode()
    else:
        return None


T = TypeVar("T")


class ApiMethods(ABC):
    @abstractmethod
    async def _request(
        self,
        http_method: RequestMethod,
        api_method: str,
        type_: Type[T],
        **params: ParamType,
    ) -> T: ...

    @abstractmethod
    async def _safe_request(
        self,
        http_method: RequestMethod,
        api_method: str,
        chat_id: int | str,
        type_: Type[T],
        **params: ParamType,
    ) -> T: ...

    async def get_updates(
        self,
        offset: int | None = None,
        limit: int | None = None,
        timeout: int | None = None,
        allowed_updates: Sequence[UpdateType] | None = None,
    ) -> tuple[Update, ...]:
        api_logger.debug(
            "Get updates offset: %r, limit: %r, timeout: %r, "
            "allowed_updates: %r",
            offset,
            limit,
            timeout,
            allowed_updates,
        )
        return await self._request(
            RequestMethod.GET,
            "getUpdates",
            tuple[Update, ...],
            offset=offset,
            limit=limit,
            timeout=timeout,
            allowed_updates=_encode_json(allowed_updates),
        )

    async def set_webhook(
        self,
        url: str | None = None,
        certificate: InputFile | None = None,
        ip_address: str | None = None,
        max_connections: int | None = None,
        allowed_updates: Sequence[UpdateType] | None = None,
        drop_pending_updates: bool | None = None,
        secret_token: str | None = None,
    ) -> bool:
        api_logger.debug("Set webhook")
        return await self._request(
            RequestMethod.POST,
            "setWebhook",
            bool,
            url=url,
            certificate=certificate,
            ip_address=ip_address,
            max_connections=max_connections,
            allowed_updates=_encode_json(allowed_updates),
            drop_pending_updates=drop_pending_updates,
            secret_token=secret_token,
        )

    async def delete_webhook(
        self,
        drop_pending_updates: bool | None = None,
    ) -> bool:
        api_logger.debug("Delete webhook")
        return await self._request(
            RequestMethod.POST,
            "deleteWebhook",
            bool,
            drop_pending_updates=drop_pending_updates,
        )

    async def get_webhook_info(
        self,
    ) -> WebhookInfo:
        api_logger.debug("Get webhook info")
        return await self._request(
            RequestMethod.GET,
            "getWebhookInfo",
            WebhookInfo,
        )

    async def get_me(
        self,
    ) -> User:
        api_logger.debug("Get me")
        return await self._request(
            RequestMethod.GET,
            "getMe",
            User,
        )

    async def log_out(
        self,
    ) -> bool:
        api_logger.debug("Log out")
        return await self._request(
            RequestMethod.POST,
            "logOut",
            bool,
        )

    async def close(
        self,
    ) -> bool:
        api_logger.debug("Close")
        return await self._request(
            RequestMethod.POST,
            "close",
            bool,
        )

    async def send_message(
        self,
        chat_id: int | str,
        text: str,
        message_thread_id: int | None = None,
        parse_mode: ParseMode | None = None,
        entities: Sequence[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send message %r to chat "%s"',
            text,
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendMessage",
            chat_id,
            Message,
            text=text,
            message_thread_id=message_thread_id,
            parse_mode=parse_mode,
            entities=_encode_json(entities),
            link_preview_options=_encode_json(link_preview_options),
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def forward_message(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_id: int,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
    ) -> Message:
        api_logger.debug(
            'Forward message %s to "%s" from "%s"',
            message_id,
            chat_id,
            from_chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "forwardMessage",
            chat_id,
            Message,
            from_chat_id=from_chat_id,
            message_id=message_id,
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
        )

    async def forward_messages(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_ids: Sequence[int],
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
    ) -> Message:
        api_logger.debug(
            'Forward messages from "%s" to "%s"',
            from_chat_id,
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "forwardMessages",
            chat_id,
            Message,
            from_chat_id=from_chat_id,
            message_ids=_encode_json(sorted(message_ids)),
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
        )

    async def copy_message(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_id: int,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> MessageId:
        api_logger.debug(
            'Copy message %s to "%s" from "%s"',
            message_id,
            chat_id,
            from_chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "copyMessage",
            chat_id,
            MessageId,
            from_chat_id=from_chat_id,
            message_id=message_id,
            message_thread_id=message_thread_id,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            caption=caption,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def copy_messages(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_ids: Sequence[int],
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        remove_caption: bool | None = None,
    ) -> MessageId:
        api_logger.debug(
            'Copy messages from "%s" to "%s"',
            from_chat_id,
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "copyMessages",
            chat_id,
            MessageId,
            from_chat_id=from_chat_id,
            message_ids=_encode_json(sorted(message_ids)),
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            remove_caption=remove_caption,
        )

    async def send_photo(
        self,
        chat_id: int | str,
        photo: InputFile | str,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send photo to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendPhoto",
            chat_id,
            Message,
            photo=photo,
            message_thread_id=message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_audio(
        self,
        chat_id: int | str,
        audio: InputFile | str,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        duration: int | None = None,
        performer: str | None = None,
        title: str | None = None,
        thumbnail: InputFile | str | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send audio to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendAudio",
            chat_id,
            Message,
            audio=audio,
            message_thread_id=message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            disable_notification=disable_notification,
            protect_content=protect_content,
            duration=duration,
            performer=performer,
            title=title,
            thumbnail=thumbnail,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_document(
        self,
        chat_id: int | str,
        document: InputFile | str,
        message_thread_id: int | None = None,
        thumbnail: InputFile | str | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        disable_content_type_detection: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        duration: int | None = None,
        performer: str | None = None,
        title: str | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send document to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendDocument",
            chat_id,
            Message,
            document=document,
            message_thread_id=message_thread_id,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            disable_content_type_detection=disable_content_type_detection,
            disable_notification=disable_notification,
            protect_content=protect_content,
            duration=duration,
            performer=performer,
            title=title,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_video(
        self,
        chat_id: int | str,
        video: InputFile | str,
        message_thread_id: int | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        has_spoiler: bool | None = None,
        supports_streaming: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send video to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendVideo",
            chat_id,
            Message,
            video=video,
            message_thread_id=message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            has_spoiler=has_spoiler,
            supports_streaming=supports_streaming,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_animation(
        self,
        chat_id: int | str,
        animation: InputFile | str,
        message_thread_id: int | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send animation to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendAnimation",
            chat_id,
            Message,
            animation=animation,
            message_thread_id=message_thread_id,
            duration=duration,
            width=width,
            height=height,
            thumbnail=thumbnail,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            has_spoiler=has_spoiler,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_voice(
        self,
        chat_id: int | str,
        voice: InputFile | str,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        duration: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send voice to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendVoice",
            chat_id,
            Message,
            voice=voice,
            message_thread_id=message_thread_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            duration=duration,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_video_note(
        self,
        chat_id: int | str,
        video_note: InputFile | str,
        message_thread_id: int | None = None,
        duration: int | None = None,
        length: int | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send video not to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendVideoNote",
            chat_id,
            Message,
            video_note=video_note,
            message_thread_id=message_thread_id,
            duration=duration,
            length=length,
            thumbnail=thumbnail,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_media_group(
        self,
        chat_id: int | str,
        media: Iterable[InputMedia],
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
    ) -> tuple[Message, ...]:
        api_logger.debug(
            'Send media group to "%s"',
            chat_id,
        )
        attached_media = []
        attachments = {}
        counter = count()
        for item in media:
            if isinstance(item.media, InputFile):
                attachment_name = f"attachment{next(counter)}"
                attachments[attachment_name] = item.media
                item = msgspec.structs.replace(
                    item, media=f"attach://{attachment_name}"
                )

            if isinstance(item, InputMediaWithThumbnail) and isinstance(
                item.thumbnail, InputFile
            ):
                attachment_name = f"attachment{next(counter)}"
                attachments[attachment_name] = item.thumbnail
                item = msgspec.structs.replace(
                    item, thumbnail=f"attach://{attachment_name}"
                )
            attached_media.append(item)
        return await self._safe_request(
            RequestMethod.POST,
            "sendMediaGroup",
            chat_id,
            tuple[Message, ...],
            media=_encode_json(attached_media),
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=_encode_json(protect_content),
            reply_parameters=_encode_json(reply_parameters),
            **attachments,
        )

    async def send_location(
        self,
        chat_id: int | str,
        latitude: float,
        longitude: float,
        message_thread_id: int | None = None,
        horizontal_accuracy: float | None = None,
        live_period: int | None = None,
        heading: int | None = None,
        proximity_alert_radius: int | None = None,
        length: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send location to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendLocation",
            chat_id,
            Message,
            latitude=latitude,
            longitude=longitude,
            message_thread_id=message_thread_id,
            horizontal_accuracy=horizontal_accuracy,
            live_period=live_period,
            heading=heading,
            proximity_alert_radius=proximity_alert_radius,
            length=length,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def edit_message_live_location(
        self,
        latitude: float,
        longitude: float,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        proximity_alert_radius: int | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message | bool:
        api_logger.debug(
            'Edit live location %s in "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "editMessageLiveLocation",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            latitude=latitude,
            longitude=longitude,
            horizontal_accuracy=horizontal_accuracy,
            heading=heading,
            proximity_alert_radius=proximity_alert_radius,
            reply_markup=_encode_json(reply_markup),
        )

    async def stop_message_live_location(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message | bool:
        api_logger.debug(
            'Stop live location %s in "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "stopMessageLiveLocation",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            reply_markup=_encode_json(reply_markup),
        )

    async def send_venue(
        self,
        chat_id: int | str,
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        message_thread_id: int | None = None,
        foursquare_id: str | None = None,
        foursquare_type: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send venue to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendVenue",
            chat_id,
            Message,
            latitude=latitude,
            longitude=longitude,
            title=title,
            address=address,
            message_thread_id=message_thread_id,
            foursquare_id=foursquare_id,
            foursquare_type=foursquare_type,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_contact(
        self,
        chat_id: int | str,
        phone_number: str,
        first_name: str,
        message_thread_id: int | None = None,
        last_name: str | None = None,
        vcard: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send contact to "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendContact",
            chat_id,
            Message,
            phone_number=phone_number,
            first_name=first_name,
            message_thread_id=message_thread_id,
            last_name=last_name,
            vcard=vcard,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_poll(
        self,
        chat_id: int | str,
        question: str,
        options: Sequence[str],
        message_thread_id: int | None = None,
        is_anonymous: bool | None = None,
        type_: PollType | None = None,
        allows_multiple_answers: bool | None = None,
        correct_option_id: int | None = None,
        is_closed: bool | None = None,
        explanation: str | None = None,
        explanation_parse_mode: ParseMode | None = None,
        explanation_entities: Sequence[MessageEntity] | None = None,
        open_period: int | None = None,
        close_date: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send poll to chat "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendPoll",
            chat_id,
            Message,
            question=question,
            options=_encode_json(options),
            message_thread_id=message_thread_id,
            is_anonymous=is_anonymous,
            type=type_,
            allows_multiple_answers=allows_multiple_answers,
            correct_option_id=correct_option_id,
            is_closed=is_closed,
            explanation=explanation,
            explanation_parse_mode=explanation_parse_mode,
            explanation_entities=_encode_json(explanation_entities),
            open_period=open_period,
            close_date=close_date,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_dice(
        self,
        chat_id: int | str,
        message_thread_id: int | None = None,
        emoji: DiceEmoji | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send dice "%s" to chat "%s"',
            emoji,
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendDice",
            chat_id,
            Message,
            message_thread_id=message_thread_id,
            emoji=emoji,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def send_chat_action(
        self,
        chat_id: int | str,
        action: ChatAction,
        message_thread_id: int | None = None,
    ) -> bool:
        api_logger.debug(
            'Send action "%s" to chat "%s"',
            action,
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendChatAction",
            chat_id,
            bool,
            action=action,
            message_thread_id=message_thread_id,
        )

    async def set_message_reaction(
        self,
        chat_id: int | str,
        message_id: int,
        reaction: Sequence[ReactionType] | None = None,
        is_big: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Send set message reaction to chat "%s"',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "setMessageReaction",
            chat_id,
            bool,
            message_id=message_id,
            reaction=_encode_json(reaction),
            is_big=is_big,
        )

    async def get_user_profile_photos(
        self,
        user_id: int,
        offset: int | None = None,
        limit: int | None = None,
    ) -> UserProfilePhotos:
        api_logger.debug(
            "Get user profile photos %s offset %s limit %s",
            user_id,
            offset,
            limit,
        )
        return await self._request(
            RequestMethod.GET,
            "getUserProfilePhotos",
            UserProfilePhotos,
            user_id=user_id,
            offset=offset,
            limit=limit,
        )

    async def get_file(
        self,
        file_id: str,
    ) -> File:
        api_logger.debug(
            'Get file "%s"',
            file_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getFile",
            File,
            file_id=file_id,
        )

    async def ban_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        until_date: int | None = None,
        revoke_messages: bool | None = None,
    ) -> bool:
        return await self._request(
            RequestMethod.POST,
            "banChatMember",
            bool,
            chat_id=chat_id,
            user_id=user_id,
            until_date=until_date,
            revoke_messages=revoke_messages,
        )

    async def unban_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        only_if_banned: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Unban member %s in "%s"',
            user_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unbanChatMember",
            bool,
            chat_id=chat_id,
            user_id=user_id,
            only_if_banned=only_if_banned,
        )

    async def restrict_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        permissions: ChatPermissions,
        use_independent_chat_permissions: bool | None = None,
        until_date: int | None = None,
    ) -> bool:
        api_logger.debug(
            'Restrict member %s in "%s"',
            user_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "restrictChatMember",
            bool,
            chat_id=chat_id,
            user_id=user_id,
            permissions=_encode_json(permissions),
            use_independent_chat_permissions=use_independent_chat_permissions,
            until_date=until_date,
        )

    async def promote_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        is_anonymous: int | None = None,
        can_manage_chat: int | None = None,
        can_change_info: int | None = None,
        can_post_messages: bool | None = None,
        can_edit_messages: bool | None = None,
        can_delete_messages: bool | None = None,
        can_manage_video_chats: bool | None = None,
        can_invite_users: bool | None = None,
        can_restrict_members: bool | None = None,
        can_pin_messages: bool | None = None,
        can_post_stories: bool | None = None,
        can_edit_stories: bool | None = None,
        can_delete_stories: bool | None = None,
        can_promote_members: bool | None = None,
        can_manage_topics: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Promote member %s in "%s"',
            user_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "promoteChatMember",
            bool,
            chat_id=chat_id,
            user_id=user_id,
            is_anonymous=is_anonymous,
            can_manage_chat=can_manage_chat,
            can_change_info=can_change_info,
            can_post_messages=can_post_messages,
            can_edit_messages=can_edit_messages,
            can_delete_messages=can_delete_messages,
            can_manage_video_chats=can_manage_video_chats,
            can_invite_users=can_invite_users,
            can_restrict_members=can_restrict_members,
            can_pin_messages=can_pin_messages,
            can_post_stories=can_post_stories,
            can_edit_stories=can_edit_stories,
            can_delete_stories=can_delete_stories,
            can_promote_members=can_promote_members,
            can_manage_topics=can_manage_topics,
        )

    async def set_chat_administrator_custom_title(
        self,
        chat_id: int | str,
        user_id: int,
        custom_title: str,
    ) -> bool:
        api_logger.debug(
            'Set title "%s" for admin %s in "%s"',
            custom_title,
            user_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setChatAdministratorCustomTitle",
            bool,
            chat_id=chat_id,
            user_id=user_id,
            custom_title=custom_title,
        )

    async def ban_chat_sender_chat(
        self,
        chat_id: int | str,
        sender_chat_id: int,
        until_date: int | None,
    ) -> bool:
        api_logger.debug(
            'Ban chat "%s" sender chat "%s"',
            chat_id,
            sender_chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "banChatSenderChat",
            bool,
            chat_id=chat_id,
            sender_chat_id=sender_chat_id,
            until_date=until_date,
        )

    async def unban_chat_sender_chat(
        self,
        chat_id: int | str,
        sender_chat_id: int,
    ) -> bool:
        api_logger.debug(
            'Unban chat "%s" sender chat "%s"',
            chat_id,
            sender_chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unbanChatSenderChat",
            bool,
            chat_id=chat_id,
            sender_chat_id=sender_chat_id,
        )

    async def export_chat_invite_link(
        self,
        chat_id: int | str,
    ) -> str:
        api_logger.debug(
            'Export chat "%s" invite link',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "exportChatInviteLink",
            str,
            chat_id=chat_id,
        )

    async def create_chat_invite_link(
        self,
        chat_id: int | str,
        name: str | None = None,
        expire_date: int | None = None,
        member_limit: int | None = None,
        creates_join_request: bool | None = None,
    ) -> ChatInviteLink:
        api_logger.debug(
            'Create chat "%s" invite link',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "createChatInviteLink",
            ChatInviteLink,
            chat_id=chat_id,
            name=name,
            expire_date=expire_date,
            member_limit=member_limit,
            creates_join_request=creates_join_request,
        )

    async def edit_chat_invite_link(
        self,
        chat_id: int | str,
        invite_link: str,
        name: str | None = None,
        expire_date: int | None = None,
        member_limit: int | None = None,
        creates_join_request: bool | None = None,
    ) -> ChatInviteLink:
        api_logger.debug(
            'Edit chat "%s" invite link',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "editChatInviteLink",
            ChatInviteLink,
            chat_id=chat_id,
            invite_link=invite_link,
            name=name,
            expire_date=expire_date,
            member_limit=member_limit,
            creates_join_request=creates_join_request,
        )

    async def revoke_chat_invite_link(
        self,
        chat_id: int | str,
        invite_link: str,
    ) -> ChatInviteLink:
        api_logger.debug(
            'Revoke chat "%s" invite link',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "revokeChatInviteLink",
            ChatInviteLink,
            chat_id=chat_id,
            invite_link=invite_link,
        )

    async def approve_chat_join_request(
        self,
        chat_id: int | str,
        user_id: int,
    ) -> bool:
        api_logger.debug(
            'Approve "%s" chat "%s" join request',
            chat_id,
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "approveChatJoinRequest",
            bool,
            chat_id=chat_id,
            user_id=user_id,
        )

    async def decline_chat_join_request(
        self,
        chat_id: int | str,
        user_id: int,
    ) -> bool:
        api_logger.debug(
            'Decline "%s" chat "%s" join request',
            chat_id,
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "declineChatJoinRequest",
            bool,
            chat_id=chat_id,
            user_id=user_id,
        )

    async def set_chat_permissions(
        self,
        chat_id: int | str,
        permissions: ChatPermissions,
        use_independent_chat_permissions: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Set chat "%s" permissions',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "setChatPermissions",
            chat_id,
            bool,
            permissions=_encode_json(permissions),
            use_independent_chat_permissions=use_independent_chat_permissions,
        )

    async def set_chat_photo(
        self,
        chat_id: int | str,
        photo: InputFile,
    ) -> bool:
        api_logger.debug(
            'Set chat "%s" photo',
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "setChatPhoto",
            chat_id,
            bool,
            photo=photo,
        )

    async def delete_chat_photo(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            'Delete chat "%s" photo',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteChatPhoto",
            bool,
            chat_id=chat_id,
        )

    async def set_chat_title(
        self,
        chat_id: int | str,
        title: str,
    ) -> bool:
        api_logger.debug(
            'Set title "%s" for chat "%s"',
            title,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setChatTitle",
            bool,
            chat_id=chat_id,
            title=title,
        )

    async def set_chat_description(
        self,
        chat_id: int | str,
        description: str,
    ) -> bool:
        api_logger.debug(
            'Set chat "%s" description',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setChatDescription",
            bool,
            chat_id=chat_id,
            description=description,
        )

    async def pin_chat_message(
        self,
        chat_id: int | str,
        message_id: int,
        disable_notification: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Pin message %s in chat "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "pinChatMessage",
            bool,
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
        )

    async def unpin_chat_message(
        self,
        chat_id: int | str,
        message_id: int | None,
    ) -> bool:
        api_logger.debug(
            'Unpin message "%s" in chat "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unpinChatMessage",
            bool,
            chat_id=chat_id,
            message_id=message_id,
        )

    async def unpin_all_chat_messages(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            'Unpin all messages in chat "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unpinAllChatMessages",
            bool,
            chat_id=chat_id,
        )

    async def leave_chat(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            'Leave chat "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "leaveChat",
            bool,
            chat_id=chat_id,
        )

    async def get_chat(
        self,
        chat_id: int | str,
    ) -> Chat:
        api_logger.debug(
            'Get chat "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getChat",
            Chat,
            chat_id=chat_id,
        )

    async def get_chat_administrators(
        self,
        chat_id: int | str,
    ) -> tuple[ChatMember, ...]:
        api_logger.debug(
            'Get chat administrators "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getChatAdministrators",
            tuple[ChatMember, ...],
            chat_id=chat_id,
        )

    async def get_chat_member_count(
        self,
        chat_id: int | str,
    ) -> int:
        api_logger.debug(
            'Get chat member count "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getChatMemberCount",
            int,
            chat_id=chat_id,
        )

    async def get_chat_member(
        self, chat_id: int | str, user_id: int
    ) -> ChatMember:
        api_logger.debug(
            'Get chat "%s" member %s',
            chat_id,
            user_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getChatMember",
            ChatMember,  # type: ignore # TODO
            chat_id=chat_id,
            user_id=user_id,
        )

    async def set_chat_sticker_set(
        self,
        chat_id: int | str,
        sticker_set_name: str,
    ) -> bool:
        api_logger.debug(
            'Set chat "%s" sticker set "%s"',
            chat_id,
            sticker_set_name,
        )
        return await self._request(
            RequestMethod.POST,
            "setChatStickerSet",
            bool,
            chat_id=chat_id,
            sticker_set_name=sticker_set_name,
        )

    async def delete_chat_sticker_set(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            'Delete chat "%s" sticker set',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteChatStickerSet",
            bool,
            chat_id=chat_id,
        )

    async def get_forum_topic_icon_stickers(
        self,
    ) -> tuple[Sticker, ...]:
        api_logger.debug(
            "Get forum topic icon stickers",
        )
        return await self._request(
            RequestMethod.GET,
            "getForumTopicIconStickers",
            tuple[Sticker, ...],
        )

    async def create_forum_topic(
        self,
        chat_id: int | str,
        name: str,
        icon_color: IconColor | None = None,
        icon_custom_emoji_id: str | None = None,
    ) -> ForumTopic:
        api_logger.debug(
            "Create forum topic %r %r",
            chat_id,
            name,
        )
        return await self._request(
            RequestMethod.POST,
            "createForumTopic",
            ForumTopic,
            chat_id=chat_id,
            icon_color=icon_color,
            icon_custom_emoji_id=icon_custom_emoji_id,
        )

    async def edit_forum_topic(
        self,
        chat_id: int | str,
        message_thread_id: int,
        name: str | None = None,
        icon_custom_emoji_id: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Edit forum topic %r %r",
            chat_id,
            message_thread_id,
        )
        return await self._request(
            RequestMethod.POST,
            "editForumTopic",
            bool,
            chat_id=chat_id,
            message_thread_id=message_thread_id,
            name=name,
            icon_custom_emoji_id=icon_custom_emoji_id,
        )

    async def close_forum_topic(
        self,
        chat_id: int | str,
        message_thread_id: int,
    ) -> bool:
        api_logger.debug(
            "Close forum topic %r %r",
            chat_id,
            message_thread_id,
        )
        return await self._request(
            RequestMethod.POST,
            "closeForumTopic",
            bool,
            chat_id=chat_id,
            message_thread_id=message_thread_id,
        )

    async def reopen_forum_topic(
        self,
        chat_id: int | str,
        message_thread_id: int,
    ) -> bool:
        api_logger.debug(
            "Reopen forum topic %r %r",
            chat_id,
            message_thread_id,
        )
        return await self._request(
            RequestMethod.POST,
            "reopenForumTopic",
            bool,
            chat_id=chat_id,
            message_thread_id=message_thread_id,
        )

    async def unpin_all_forum_topic_messages(
        self,
        chat_id: int | str,
        message_thread_id: int,
    ) -> bool:
        api_logger.debug(
            "Unpin all forum topic messages %r %r",
            chat_id,
            message_thread_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unpinAllForumTopicMessages",
            bool,
            chat_id=chat_id,
            message_thread_id=message_thread_id,
        )

    async def edit_general_forum_topic(
        self,
        chat_id: int | str,
        name: str,
    ) -> bool:
        api_logger.debug(
            "Edit general forum topic %r %r",
            chat_id,
            name,
        )
        return await self._request(
            RequestMethod.POST,
            "editGeneralForumTopic",
            bool,
            chat_id=chat_id,
            name=name,
        )

    async def close_general_forum_topic(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            "Close general forum topic %r",
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "closeGeneralForumTopic",
            bool,
            chat_id=chat_id,
        )

    async def reopen_general_forum_topic(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            "Reopen general forum topic %r",
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "reopenGeneralForumTopic",
            bool,
            chat_id=chat_id,
        )

    async def hide_general_forum_topic(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            "Hide general forum topic %r",
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "hideGeneralForumTopic",
            bool,
            chat_id=chat_id,
        )

    async def unhide_general_forum_topic(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            "Unhide general forum topic %r",
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unhideGeneralForumTopic",
            bool,
            chat_id=chat_id,
        )

    async def unpin_all_general_forum_topic_messages(
        self,
        chat_id: int | str,
    ) -> bool:
        api_logger.debug(
            "Unpin all general forum topic messages %r",
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "unpinAllGeneralForumTopicMessages",
            bool,
            chat_id=chat_id,
        )

    async def answer_callback_query(
        self,
        callback_query_id: str,
        text: str | None = None,
        show_alert: bool | None = None,
        url: str | None = None,
        cache_time: int | None = None,
    ) -> bool:
        api_logger.debug(
            'Answer callback query "%s"',
            callback_query_id,
        )
        return await self._request(
            RequestMethod.POST,
            "answerCallbackQuery",
            bool,
            callback_query_id=callback_query_id,
            text=text,
            show_alert=show_alert,
            url=url,
            cache_time=cache_time,
        )

    async def get_user_chat_boosts(
        self,
        chat_id: int | str,
        user_id: int,
    ) -> UserChatBoosts:
        api_logger.debug(
            "Get user chat boosts %r %r",
            chat_id,
            user_id,
        )
        return await self._request(
            RequestMethod.GET,
            "getUserChatBoosts",
            UserChatBoosts,
            chat_id=chat_id,
            user_id=user_id,
        )

    async def set_my_commands(
        self,
        commands: Sequence[BotCommand],
        scope: BotCommandScope | None = None,
        language_code: str | None = None,
    ) -> bool:
        api_logger.debug(
            'Set my commands "%s"',
            commands,
        )
        return await self._request(
            RequestMethod.POST,
            "setMyCommands",
            bool,
            commands=_encode_json(commands),
            scope=_encode_json(scope),
            language_code=language_code,
        )

    async def set_chat_menu_button(
        self,
        chat_id: int | None,
        menu_button: MenuButton | None,
    ) -> bool:
        api_logger.debug(
            'Set chat menu button "%r"',
            menu_button,
        )
        return await self._request(
            RequestMethod.POST,
            "setChatMenuButton",
            bool,
            chat_id=chat_id,
            menu_button=_encode_json(menu_button),
        )

    async def get_chat_menu_button(
        self,
        chat_id: int | None,
    ) -> MenuButton:
        api_logger.debug(
            'Get chat menu button "%r"',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "getChatMenuButton",
            MenuButton,
            chat_id=chat_id,
        )

    async def set_my_default_administrator_rights(
        self,
        rights: ChatAdministratorRights | None,
        for_channels: bool | None,
    ) -> bool:
        api_logger.debug(
            "Set my default administrator rights %r",
            rights,
        )
        return await self._request(
            RequestMethod.POST,
            "setMyDefaultAdministratorRights",
            bool,
            rights=_encode_json(rights),
            for_channels=for_channels,
        )

    async def get_my_default_administrator_rights(
        self,
        for_channels: bool | None,
    ) -> ChatAdministratorRights:
        api_logger.debug(
            "Get my default administrator rights",
        )
        return await self._request(
            RequestMethod.POST,
            "getMyDefaultAdministratorRights",
            ChatAdministratorRights,
            for_channels=for_channels,
        )

    async def delete_my_commands(
        self,
        scope: BotCommandScope | None = None,
        language_code: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Delete my commands",
        )
        return await self._request(
            RequestMethod.GET,
            "deleteMyCommands",
            bool,
            scope=_encode_json(scope),
            language_code=language_code,
        )

    async def get_my_commands(
        self,
        scope: BotCommandScope | None = None,
        language_code: str | None = None,
    ) -> tuple[BotCommand, ...]:
        api_logger.debug(
            "Get my commands",
        )
        return await self._request(
            RequestMethod.GET,
            "getMyCommands",
            tuple[BotCommand, ...],
            scope=_encode_json(scope),
            language_code=language_code,
        )

    async def set_my_name(
        self,
        name: str | None = None,
        language_code: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Set my name",
        )
        return await self._request(
            RequestMethod.POST,
            "setMyName",
            bool,
            name=name,
            language_code=language_code,
        )

    async def get_my_name(
        self,
        language_code: str | None = None,
    ) -> BotName:
        api_logger.debug(
            "Get my name",
        )
        return await self._request(
            RequestMethod.GET,
            "getMyName",
            BotName,
            language_code=language_code,
        )

    async def set_my_description(
        self,
        description: str | None = None,
        language_code: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Set my description",
        )
        return await self._request(
            RequestMethod.POST,
            "setMyDescription",
            bool,
            description=description,
            language_code=language_code,
        )

    async def get_my_description(
        self,
        language_code: str | None = None,
    ) -> BotDescription:
        api_logger.debug(
            "Get my description",
        )
        return await self._request(
            RequestMethod.GET,
            "getMyDescription",
            BotDescription,
            language_code=language_code,
        )

    async def set_my_short_description(
        self,
        short_description: str | None = None,
        language_code: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Set my short description",
        )
        return await self._request(
            RequestMethod.POST,
            "setMyShortDescription",
            bool,
            short_description=short_description,
            language_code=language_code,
        )

    async def get_my_short_description(
        self,
        language_code: str | None = None,
    ) -> BotShortDescription:
        api_logger.debug(
            "Get my short description",
        )
        return await self._request(
            RequestMethod.GET,
            "getMyShortDescription",
            BotShortDescription,
            language_code=language_code,
        )

    async def edit_message_text(
        self,
        text: str,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        parse_mode: ParseMode | None = None,
        entities: Sequence[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        if (
            chat_id is None or message_id is None
        ) and inline_message_id is None:
            raise RuntimeError(
                "chat_id or message_id and inline_message_id is None"
            )
        if inline_message_id is None:
            api_logger.debug(
                'Edit message %s in "%s" text',
                message_id,
                chat_id,
            )
        else:
            api_logger.debug(
                'Edit inline message "%s" text',
                inline_message_id,
            )
        return await self._request(
            RequestMethod.POST,
            "editMessageText",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            text=text,
            parse_mode=parse_mode,
            entities=_encode_json(entities),
            link_preview_options=_encode_json(link_preview_options),
            reply_markup=_encode_json(reply_markup),
        )

    async def edit_message_caption(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        caption: str | None = None,
        parse_mode: ParseMode | None = None,
        caption_entities: Sequence[MessageEntity] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        if (
            chat_id is None or message_id is None
        ) and inline_message_id is None:
            raise RuntimeError(
                "chat_id or message_id and inline_message_id is None"
            )
        if inline_message_id is None:
            api_logger.debug(
                'Edit message %s in "%s" caption',
                message_id,
                chat_id,
            )
        else:
            api_logger.debug(
                'Edit inline message "%s" caption',
                inline_message_id,
            )
        return await self._request(
            RequestMethod.POST,
            "editMessageCaption",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            caption=caption,
            parse_mode=parse_mode,
            caption_entities=_encode_json(caption_entities),
            reply_markup=_encode_json(reply_markup),
        )

    async def edit_message_media(
        self,
        media: InputMedia,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        if (
            chat_id is None or message_id is None
        ) and inline_message_id is None:
            raise RuntimeError(
                "chat_id or message_id and inline_message_id is None"
            )
        if inline_message_id is None:
            api_logger.debug(
                'Edit message %s in "%s" media',
                message_id,
                chat_id,
            )
        else:
            api_logger.debug(
                'Edit inline message "%s" media',
                inline_message_id,
            )
        attachments = {}
        if isinstance(media.media, InputFile):
            attachment_name = "attachment0"
            attachments[attachment_name] = media.media
            media = msgspec.structs.replace(
                media, media=f"attach://{attachment_name}"
            )
        if isinstance(media, InputMediaWithThumbnail) and isinstance(
            media.thumbnail, InputFile
        ):
            attachment_name = "attachment1"
            attachments[attachment_name] = media.thumbnail
            media = msgspec.structs.replace(
                media, thumbnail=f"attach://{attachment_name}"
            )
        return await self._request(
            RequestMethod.POST,
            "editMessageMedia",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            media=_encode_json(media),
            reply_markup=_encode_json(reply_markup),
            **attachments,
        )

    async def edit_message_reply_markup(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message | bool:
        if (
            chat_id is None or message_id is None
        ) and inline_message_id is None:
            raise RuntimeError(
                "chat_id or message_id and inline_message_id is None"
            )
        if inline_message_id is None:
            api_logger.debug(
                'Edit message %s in "%s" reply markup',
                message_id,
                chat_id,
            )
        else:
            api_logger.debug(
                'Edit inline message "%s" reply markup',
                inline_message_id,
            )
        return await self._request(
            RequestMethod.POST,
            "editMessageReplyMarkup",
            Message | bool,  # type: ignore # TODO
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            reply_markup=_encode_json(reply_markup),
        )

    async def stop_poll(
        self,
        chat_id: int | str,
        message_id: int,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Poll:
        api_logger.debug(
            'Stop poll %s in "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "stopPoll",
            Poll,
            chat_id=chat_id,
            message_id=message_id,
            reply_markup=_encode_json(reply_markup),
        )

    async def delete_message(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
    ) -> bool:
        api_logger.debug(
            'Delete message %s in "%s"',
            message_id,
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteMessage",
            bool,
            chat_id=chat_id,
            message_id=message_id,
        )

    async def delete_messages(
        self,
        chat_id: int | str,
        message_ids: Sequence[int],
    ) -> bool:
        api_logger.debug(
            'Delete messages in "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteMessages",
            bool,
            chat_id=chat_id,
            message_ids=_encode_json(sorted(message_ids)),
        )

    async def send_sticker(
        self,
        chat_id: int | str,
        sticker: InputFile | str,
        emoji: str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: ReplyMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send sticker to "%s"',
            chat_id,
        )
        return await self._request(
            RequestMethod.POST,
            "sendSticker",
            Message,
            chat_id=chat_id,
            sticker=sticker,
            emoji=emoji,
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def get_sticker_set(
        self,
        name: str,
    ) -> StickerSet:
        api_logger.debug(
            'Get sticker set "%s"',
            name,
        )
        return await self._request(
            RequestMethod.GET,
            "getStickerSet",
            StickerSet,
            name=name,
        )

    async def get_custom_emoji_stickers(
        self,
        custom_emoji_ids: Sequence[str],
    ) -> tuple[Sticker, ...]:
        api_logger.debug("Get custom emoji stickers")
        return await self._request(
            RequestMethod.GET,
            "getCustomEmojiStickers",
            tuple[Sticker, ...],
            custom_emoji_ids=_encode_json(custom_emoji_ids),
        )

    async def upload_sticker_file(
        self,
        user_id: int,
        sticker: InputFile,
        sticker_format: StickerFormat,
    ) -> File:
        api_logger.debug(
            "Upload sticker file for %s",
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "uploadStickerFile",
            File,
            user_id=user_id,
            sticker=sticker,
            sticker_format=sticker_format,
        )

    async def create_new_sticker_set(
        self,
        user_id: int,
        name: str,
        title: str,
        stickers: Iterable[InputSticker],
        sticker_format: StickerFormat,
        sticker_type: StickerType | None = None,
        needs_repainting: bool | None = None,
    ) -> bool:
        api_logger.debug(
            'Create new sticker set "%s" for %s',
            name,
            user_id,
        )
        attached_media = []
        attachments = {}
        counter = count()
        for sticker in stickers:
            if isinstance(sticker.sticker, str):
                attached_media.append(sticker)
            else:
                attachment_name = f"attachment{next(counter)}"
                attachments[attachment_name] = sticker.sticker
                attached_media.append(
                    msgspec.structs.replace(
                        sticker, sticker=f"attach://{attachment_name}"
                    )
                )
        return await self._request(
            RequestMethod.POST,
            "createNewStickerSet",
            bool,
            user_id=user_id,
            name=name,
            title=title,
            stickers=_encode_json(attached_media),
            sticker_format=sticker_format,
            sticker_type=sticker_type,
            needs_repainting=needs_repainting,
        )

    async def add_sticker_to_set(
        self,
        user_id: int,
        name: str,
        sticker: InputSticker,
    ) -> bool:
        api_logger.debug(
            'Add sticker to set "%s" for %s',
            name,
            user_id,
        )
        attachments = {}
        if not isinstance(sticker.sticker, str):
            attachment_name = "attachment0"
            attachments[attachment_name] = sticker.sticker
            sticker = msgspec.structs.replace(
                sticker, media=f"attach://{attachment_name}"
            )
        return await self._request(
            RequestMethod.POST,
            "addStickerToSet",
            bool,
            user_id=user_id,
            name=name,
            sticker=_encode_json(sticker),
            **attachments,
        )

    async def set_sticker_position_in_set(
        self,
        sticker: str,
        position: int,
    ) -> bool:
        api_logger.debug(
            'Set sticker "%s" position to %s',
            sticker,
            position,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerPositionInSet",
            bool,
            sticker=sticker,
            position=position,
        )

    async def delete_sticker_from_set(
        self,
        sticker: str,
    ) -> bool:
        api_logger.debug(
            'Delete sticker "%s" from set',
            sticker,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteStickerFromSet",
            bool,
            sticker=sticker,
        )

    async def set_sticker_emoji_list(
        self,
        sticker: str,
        emoji_list: Sequence[str],
    ) -> bool:
        api_logger.debug(
            "Set sticker emoji list %r",
            sticker,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerEmojiList",
            bool,
            sticker=sticker,
            emoji_list=_encode_json(emoji_list),
        )

    async def set_sticker_keywords(
        self,
        sticker: str,
        keywords: Sequence[str],
    ) -> bool:
        api_logger.debug(
            "Set sticker keywords %r",
            sticker,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerKeywords",
            bool,
            sticker=sticker,
            emoji_list=_encode_json(keywords),
        )

    async def set_sticker_mask_position(
        self,
        sticker: str,
        mask_position: MaskPosition,
    ) -> bool:
        api_logger.debug(
            "Set sticker mask position %r",
            sticker,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerMaskPosition",
            bool,
            sticker=sticker,
            mask_position=_encode_json(mask_position),
        )

    async def set_sticker_set_title(
        self,
        name: str,
        title: str,
    ) -> bool:
        api_logger.debug(
            "Set sticker set title %r %r",
            name,
            title,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerSetTitle",
            bool,
            name=name,
            title=title,
        )

    async def set_sticker_set_thumbnail(
        self,
        name: str,
        user_id: int,
        thumbnail: InputFile | str | None = None,
    ) -> bool:
        api_logger.debug(
            "Set sticker set thumbnail %r %r",
            name,
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setStickerSetThumbnail",
            bool,
            name=name,
            user_id=user_id,
            thumbnail=thumbnail,
        )

    async def set_custom_emoji_sticker_set_thumbnail(
        self,
        name: str,
        custom_emoji_id: str | None = None,
    ) -> bool:
        api_logger.debug(
            "Set custom emoji sticker set thumbnail %r %r",
            name,
            custom_emoji_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setCustomEmojiStickerSetThumbnail",
            bool,
            name=name,
            custom_emoji_id=custom_emoji_id,
        )

    async def delete_sticker_set(
        self,
        name: str,
    ) -> bool:
        api_logger.debug(
            "Delete sticker set %r",
            name,
        )
        return await self._request(
            RequestMethod.POST,
            "deleteStickerSet",
            bool,
            name=name,
        )

    async def answer_inline_query(
        self,
        inline_query_id: str,
        results: Sequence[InlineQueryResult],
        cache_time: int | None = None,
        is_personal: bool | None = None,
        next_offset: str | None = None,
        button: InlineQueryResultsButton | None = None,
    ) -> bool:
        api_logger.debug(
            'Answer inline query "%s"',
            inline_query_id,
        )
        return await self._request(
            RequestMethod.POST,
            "answerInlineQuery",
            bool,
            inline_query_id=inline_query_id,
            results=_encode_json(results),
            cache_time=cache_time,
            is_personal=is_personal,
            next_offset=next_offset,
            button=_encode_json(button),
        )

    async def answer_web_app_query(
        self,
        web_app_query_id: str,
        results: Sequence[InlineQueryResult],
    ) -> SentWebAppMessage:
        api_logger.debug(
            'Answer web app query "%s"',
            web_app_query_id,
        )
        return await self._request(
            RequestMethod.POST,
            "answerWebAppQuery",
            SentWebAppMessage,
            web_app_query_id=web_app_query_id,
            results=_encode_json(results),
        )

    async def send_invoice(
        self,
        chat_id: int,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: Sequence[LabeledPrice],
        message_thread_id: int | None = None,
        max_tip_amount: int | None = None,
        suggested_tip_amounts: tuple[int, ...] | None = None,
        start_parameter: str | None = None,
        provider_data: str | None = None,
        photo_url: str | None = None,
        photo_size: int | None = None,
        photo_width: int | None = None,
        photo_height: int | None = None,
        need_name: bool | None = None,
        need_phone_number: bool | None = None,
        need_email: bool | None = None,
        need_shipping_address: bool | None = None,
        send_phone_number_to_provider: bool | None = None,
        send_email_to_provider: bool | None = None,
        is_flexible: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            "Send invoice to %s",
            chat_id,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendInvoice",
            chat_id,
            Message,
            title=title,
            description=description,
            payload=payload,
            provider_token=provider_token,
            currency=currency,
            prices=_encode_json(prices),
            message_thread_id=message_thread_id,
            max_tip_amount=max_tip_amount,
            suggested_tip_amounts=_encode_json(suggested_tip_amounts),
            start_parameter=start_parameter,
            provider_data=provider_data,
            photo_url=photo_url,
            photo_size=photo_size,
            photo_width=photo_width,
            photo_height=photo_height,
            need_name=need_name,
            need_phone_number=need_phone_number,
            need_email=need_email,
            need_shipping_address=need_shipping_address,
            send_phone_number_to_provider=send_phone_number_to_provider,
            send_email_to_provider=send_email_to_provider,
            is_flexible=is_flexible,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def create_invoice_link(
        self,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: Sequence[LabeledPrice],
        max_tip_amount: int | None = None,
        suggested_tip_amounts: tuple[int, ...] | None = None,
        start_parameter: str | None = None,
        provider_data: str | None = None,
        photo_url: str | None = None,
        photo_size: int | None = None,
        photo_width: int | None = None,
        photo_height: int | None = None,
        need_name: bool | None = None,
        need_phone_number: bool | None = None,
        need_email: bool | None = None,
        need_shipping_address: bool | None = None,
        send_phone_number_to_provider: bool | None = None,
        send_email_to_provider: bool | None = None,
        is_flexible: bool | None = None,
    ) -> str:
        api_logger.debug(
            "Create invoice link",
        )
        return await self._request(
            RequestMethod.POST,
            "createInvoiceLink",
            str,
            title=title,
            description=description,
            payload=payload,
            provider_token=provider_token,
            currency=currency,
            prices=_encode_json(prices),
            max_tip_amount=max_tip_amount,
            suggested_tip_amounts=_encode_json(suggested_tip_amounts),
            start_parameter=start_parameter,
            provider_data=provider_data,
            photo_url=photo_url,
            photo_size=photo_size,
            photo_width=photo_width,
            photo_height=photo_height,
            need_name=need_name,
            need_phone_number=need_phone_number,
            need_email=need_email,
            need_shipping_address=need_shipping_address,
            send_phone_number_to_provider=send_phone_number_to_provider,
            send_email_to_provider=send_email_to_provider,
            is_flexible=is_flexible,
        )

    async def answer_shipping_query(
        self,
        inline_query_id: str,
        ok: bool,
        shipping_options: Sequence[ShippingOption] | None = None,
        error_message: str | None = None,
    ) -> bool:
        api_logger.debug(
            'Answer shipping query "%s"',
            inline_query_id,
        )
        return await self._request(
            RequestMethod.POST,
            "answerShippingQuery",
            bool,
            inline_query_id=inline_query_id,
            ok=ok,
            shipping_options=_encode_json(shipping_options),
            error_message=error_message,
        )

    async def answer_pre_checkout_query(
        self,
        pre_checkout_query_id: str,
        ok: bool,
        error_message: str | None = None,
    ) -> bool:
        api_logger.debug(
            'Answer pre checkout query "%s"',
            pre_checkout_query_id,
        )
        return await self._request(
            RequestMethod.POST,
            "answerPreCheckoutQuery",
            bool,
            pre_checkout_query_id=pre_checkout_query_id,
            ok=ok,
            error_message=error_message,
        )

    async def set_passport_data_errors(
        self,
        user_id: int,
        errors: Sequence[PassportElementError],
    ) -> bool:
        api_logger.debug(
            "Set passport data errors %s",
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setPassportDataErrors",
            bool,
            user_id=user_id,
            errors=_encode_json(errors),
        )

    async def send_game(
        self,
        chat_id: int,
        game_short_name: str,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
    ) -> Message:
        api_logger.debug(
            'Send game "%s" to %s',
            chat_id,
            game_short_name,
        )
        return await self._safe_request(
            RequestMethod.POST,
            "sendGame",
            chat_id,
            Message,
            game_short_name=game_short_name,
            message_thread_id=message_thread_id,
            disable_notification=disable_notification,
            protect_content=protect_content,
            reply_parameters=_encode_json(reply_parameters),
            reply_markup=_encode_json(reply_markup),
        )

    async def set_game_score(
        self,
        user_id: int,
        score: int,
        force: bool | None = None,
        disable_edit_message: bool | None = None,
        chat_id: int | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
    ) -> Message | bool:
        api_logger.debug(
            "Set game score %s for %s",
            score,
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "setGameScore",
            Message | bool,  # type: ignore # TODO
            user_id=user_id,
            score=score,
            force=force,
            disable_edit_message=disable_edit_message,
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
        )

    async def get_game_high_scores(
        self,
        user_id: int,
        chat_id: int | None = None,
        message_id: int | None = None,
        inline_message_id: str | None = None,
    ) -> tuple[GameHighScore, ...]:
        api_logger.debug(
            "Get game high scores for %s",
            user_id,
        )
        return await self._request(
            RequestMethod.POST,
            "getGameHighScores",
            tuple[GameHighScore, ...],
            user_id=user_id,
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
        )



import re
from dataclasses import dataclass
from typing import Final

from .bot import Bot, FilterProtocol
from .bot_update import BotUpdate
from .constants import ChatType, ContentType, UpdateType

__all__ = (
    "ANDFilter",
    "CallbackQueryDataFilter",
    "CommandsFilter",
    "ContentTypeFilter",
    "GroupChatFilter",
    "MessageTextFilter",
    "ORFilter",
    "PrivateChatFilter",
    "StateFilter",
    "UpdateTypeFilter",
)


@dataclass(frozen=True)
class UpdateTypeFilter:
    update_type: UpdateType

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        return getattr(update, self.update_type) is not None


@dataclass(frozen=True)
class StateFilter:
    state: str

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        return update.state == self.state


@dataclass(frozen=True)
class CommandsFilter:
    commands: tuple[str, ...]

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        if update.message is None or update.message.text is None:
            return False
        if any(
            update.message.text.startswith(f"/{command}")
            for command in self.commands
        ):
            return True
        return False


@dataclass(frozen=True)
class ContentTypeFilter:
    content_types: tuple[ContentType, ...]

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        if update.message is not None:
            message = update.message
        elif update.edited_message is not None:
            message = update.edited_message
        elif update.channel_post is not None:
            message = update.channel_post
        elif update.edited_channel_post is not None:
            message = update.edited_channel_post
        else:
            return False
        for content_type in self.content_types:
            if getattr(message, content_type) is not None:
                return True
        return False


@dataclass(frozen=True)
class MessageTextFilter:
    pattern: "re.Pattern[str]"

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        return (
            update.message is not None
            and update.message.text is not None
            and self.pattern.match(update.message.text) is not None
        )


@dataclass(frozen=True)
class CallbackQueryDataFilter:
    pattern: "re.Pattern[str]"

    async def check(self, _: Bot, update: BotUpdate) -> bool:
        return (
            update.callback_query is not None
            and update.callback_query.data is not None
            and self.pattern.match(update.callback_query.data) is not None
        )


@dataclass(frozen=True)
class PrivateChatFilter:
    async def check(self, _: Bot, update: BotUpdate) -> bool:  # noqa
        return (
            update.message is not None
            and update.message.chat is not None
            and update.message.chat.type == ChatType.PRIVATE
        )


@dataclass(frozen=True)
class GroupChatFilter:
    async def check(self, _: Bot, update: BotUpdate) -> bool:  # noqa
        group_types = (ChatType.GROUP, ChatType.SUPERGROUP)
        return (
            update.message is not None
            and update.message.chat is not None
            and update.message.chat.type in group_types
        )


class ORFilter:
    def __init__(self, *filters: FilterProtocol) -> None:
        self._filters: Final[tuple[FilterProtocol, ...]] = filters

    async def check(self, bot: Bot, update: BotUpdate) -> bool:
        for filter_item in self._filters:
            if await filter_item.check(bot, update):
                return True
        return False


class ANDFilter:
    def __init__(self, *filters: FilterProtocol) -> None:
        self._filters: Final[tuple[FilterProtocol, ...]] = filters

    async def check(self, bot: Bot, update: BotUpdate) -> bool:
        for filter_item in self._filters:
            if not await filter_item.check(bot, update):
                return False
        return True



import re
from pathlib import Path

from setuptools import setup

path = Path(__file__).parent
txt = (path / "aiotgbot" / "__init__.py").read_text("utf-8")
print(txt)
version = re.findall(r"^__version__ = \"([^\"]+)\"\r?$", txt, re.M)[0]
readme = (path / "README.rst").read_text("utf-8")

setup(
    name="aiotgbot",
    version=version,
    description="Asynchronous library for Telegram bot API",
    long_description=readme,
    long_description_content_type="text/x-rst",
    url="https://github.com/gleb-chipiga/aiotgbot",
    license="MIT",
    author="Gleb Chipiga",
    # author_email='',
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Internet",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
    ],
    packages=["aiotgbot"],
    package_data={"aiotgbot": ["py.typed"]},
    python_requires=">=3.11,<3.13",
    install_requires=[
        "aiofreqlimit>=0.0.12",
        "aiohttp>=3.9",
        "aiojobs>=1.2.1",
        "backoff>=2",
        "frozenlist",
        "msgspec",
        "yarl",
    ],
    tests_require=[
        "hypothesis",
        "more-itertools",
        "pytest",
        "pytest-asyncio>=0.19",
        "pytest-cov",
        "sqlalchemy[aiosqlite]",
    ],
    extras_require={
        "sqlite": ["aiosqlite"],
        "passport": ["cryptography<=38"],
        "sqlalchemy": ["sqlalchemy"],
    },
)



from typing import Annotated, Any, AsyncIterator, Final

from sqlalchemy import JSON, Text, delete, insert, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from .helpers import Json
from .storage import StorageProtocol

__all__ = ("SqlalchemyStorage",)


class Base(DeclarativeBase):
    pass


class KV(Base):
    __tablename__ = "kv"

    key: Mapped[Annotated[str, mapped_column(Text, primary_key=True)]]
    value: Mapped[Annotated[Json, mapped_column(JSON)]]


class SqlalchemyStorage(StorageProtocol):
    def __init__(self, url: str) -> None:
        self._engine: Final = create_async_engine(url)

    async def connect(self) -> None:
        async with self._engine.begin() as connection:
            await connection.run_sync(Base.metadata.create_all)

    async def close(self) -> None:
        await self._engine.dispose()

    async def set(self, key: str, value: Json = None) -> None:
        async with self._engine.begin() as connection:
            try:
                async with connection.begin_nested():
                    await connection.execute(
                        insert(KV).values(key=key, value=value)
                    )
            except IntegrityError:
                await connection.execute(
                    update(KV).where(KV.key == key).values(value=value)
                )

    async def get(self, key: str) -> Json:
        async with self._engine.begin() as connection:
            result = await connection.execute(
                select(KV.value).where(KV.key == key)
            )
            return result.scalar()

    async def delete(self, key: str) -> None:
        async with self._engine.begin() as connection:
            await connection.execute(delete(KV).where(KV.key == key))

    async def iterate(
        self, prefix: str = ""
    ) -> AsyncIterator[tuple[str, Json]]:
        async with self._engine.begin() as connection:
            result = await connection.stream(
                select(KV.key, KV.value).where(KV.key.startswith(prefix))
            )
            async for key, value in result:
                yield key, value

    async def clear(self) -> None:
        async with self._engine.begin() as connection:
            await connection.execute(delete(KV))

    def raw_connection(self) -> Any:
        return self._engine



import asyncio
from contextlib import asynccontextmanager
from typing import AsyncIterator, Final, Hashable
from weakref import WeakValueDictionary

import msgspec.json
from aiohttp import web

__all__ = (
    "BotKey",
    "Json",
    "KeyLock",
    "get_python_version",
    "get_software",
    "json_dumps",
)

Json = str | int | float | bool | dict[str, "Json"] | list["Json"] | None


def json_dumps(obj: Json) -> str:
    return msgspec.json.encode(obj).decode()


def get_python_version() -> str:
    from sys import version_info as version

    return f"{version.major}.{version.minor}.{version.micro}"


def get_software() -> str:
    from . import __version__

    return f"Python/{get_python_version()} aiotgbot/{__version__}"


class KeyLock:
    def __init__(self) -> None:
        self._locks: "Final[WeakValueDictionary[Hashable," "asyncio.Lock]]" = (
            WeakValueDictionary()
        )

    @asynccontextmanager
    async def resource(self, key: Hashable) -> AsyncIterator[None]:
        if key not in self._locks:
            lock = asyncio.Lock()
            self._locks[key] = lock
        else:
            lock = self._locks[key]
        async with lock:
            yield


BotKey = web.AppKey
